# -*- coding: utf-8 -*-

# -----------------------------------------------------------------------------
# Copyright (c) 2016-2017 Anaconda, Inc.
#
# May be copied and distributed freely only as part of an Anaconda or
# Miniconda installation.
# -----------------------------------------------------------------------------

"""
This module provides user configuration file management features.

It's based on the ConfigParser module (present in the standard library).
"""

from __future__ import print_function

import ast
import configparser
import contextlib
import copy
import io
import os
import re
import shutil
import time
from anaconda_navigator.config.base import get_home_dir


# -----------------------------------------------------------------------------
# --- Auxiliary classes
# -----------------------------------------------------------------------------
class NoDefault:  # pylint: disable=too-few-public-methods
    """NoDefault object."""


# -----------------------------------------------------------------------------
# --- Defaults class
# -----------------------------------------------------------------------------
class DefaultsConfig(configparser.ConfigParser):
    """Class used to save defaults to a file."""

    def __init__(self, name, subfolder):
        """
        Class used to save defaults to a file.

        Parameters
        -----------
        name: str
            Name of the configuration.
        subfolder: str
            Path to folder location of configuration file.
        """
        configparser.ConfigParser.__init__(self)
        self.name = name
        self.subfolder = subfolder

    def _write(self, fp):  # pylint: disable=invalid-name
        """Private write method for Python 2.

        The one from configparser fails for non-ascii Windows accounts.
        """
        if self._defaults:
            fp.write(f'[{configparser.DEFAULTSECT}]\n')
            for (key, value) in self._defaults.items():
                write_value = str(value).replace('\n', '\n\t')
                fp.write(f'{key} = {write_value}\n')
            fp.write('\n')
        for section in self._sections:
            fp.write(f'[{section}]\n')
            for (key, value) in self._sections[section].items():
                if key == '__name__':
                    continue
                if (value is not None) or (self._optcre == self.OPTCRE):
                    value = str(value)
                    key = ' = '.join((key, value.replace('\n', '\n\t')))
                fp.write(f'{key}\n')
            fp.write('\n')

    def _set(self, section, option, value, verbose):
        """Private set method."""
        if not self.has_section(section):
            self.add_section(section)
        if not isinstance(value, str):
            value = repr(value)
        if verbose:
            print(f'{section}[ {option} ] = {value}')
        configparser.ConfigParser.set(self, section, option, value)

    def _save(self):
        """Save config into the associated .ini file."""
        fname = self.filename()

        def _write_file(fname):
            with open(fname, 'w', encoding='utf-8') as configfile:
                self.write(configfile)

        try:  # the "easy" way
            _write_file(fname)
        except IOError:
            try:  # the "delete and sleep" way
                if os.path.isfile(fname):
                    os.remove(fname)
                time.sleep(0.05)
                _write_file(fname)
            except Exception:
                print('Failed to write user configuration file.')
                print('Please submit a bug report.')
                raise

    def filename(self):
        """Create a .ini filename located in user home directory."""
        folder = os.path.join(get_home_dir(), self.subfolder)
        # Save defaults in a "defaults" dir of .anaconda_navigator to not
        # pollute it
        if 'defaults' in self.name:
            folder = os.path.join(folder, 'defaults')
        try:
            os.makedirs(folder)
        except os.error:
            # Folder (or one of its parents) already exists
            pass
        ini_file = os.path.join(folder, f'{self.name}.ini')
        return ini_file

    def set_defaults(self, defaults):
        """Set default configuration."""
        for section, options in defaults:
            for option in options:
                new_value = copy.deepcopy(options[option])
                self._set(section, option, new_value, False)


# -----------------------------------------------------------------------------
# --- User config class
# -----------------------------------------------------------------------------
class UserConfig(DefaultsConfig):
    """
    User configuration class.

    Note that 'get' and 'set' arguments number and type differ from the
    overriden methods.
    """

    DEFAULT_SECTION_NAME = 'main'  # pylint: disable=invalid-name

    def __init__(  # pylint: disable=too-many-arguments
        self,
        name,
        defaults=None,
        load=True,
        version=None,
        subfolder=None,
        backup=False,
        raw_mode=False,
        remove_obsolete=False
    ):
        """
        User configuration class.

        Parameters
        ----------
        name: str
            Name of the configuration.
        defaults: dict
            dictionnary containing options *or* list of tuples (section_name,
            options).
        version: str
            Version of the configuration file (X.Y.Z format).
        subfolder: str
            Configuration file will be saved in HOME/subfolder/name.ini.
        """
        DefaultsConfig.__init__(self, name, subfolder)

        self.raw = 1 if raw_mode else 0
        if (version is not None) and (re.match(r'^(\d+).(\d+).(\d+)$', version) is None):
            raise ValueError(f'Version number {version} is incorrect - must be in X.Y.Z format')

        if isinstance(defaults, dict):
            defaults = [(self.DEFAULT_SECTION_NAME, defaults)]
        self.defaults = defaults
        if defaults is not None:
            self.reset_to_defaults(save=False)

        fname = self.filename()
        if backup:
            try:
                shutil.copyfile(fname, f'{fname}.bak')
            except IOError:
                pass

        if load:
            # If config file already exists, it overrides Default options:
            self.load_from_ini()
            old_ver = self.get_version(version)

            # Save new defaults
            self.__save_new_defaults(defaults, version, subfolder)

            # Updating defaults only if major/minor version is different
            if self._minor(version) != self._minor(old_ver):
                if backup:
                    try:
                        shutil.copyfile(fname, f'{fname}-{old_ver}.bak')
                    except IOError:
                        pass
                self.__update_defaults(defaults, old_ver)

                # Remove deprecated options if major version has changed
                if (remove_obsolete or self._major(version) != self._major(old_ver)):
                    self.__remove_deprecated_options(old_ver)

                # Set new version number
                self.set_version(version, save=False)
            if defaults is None:
                # If no defaults are defined, set .ini file settings as default
                self.set_as_defaults()

    @staticmethod
    def _major(_t):
        """Return major component in config versions."""
        return _t[:_t.find('.')]

    @staticmethod
    def _minor(_t):
        """Return minor component in config versions."""
        return _t[:_t.rfind('.')]

    def get_version(self, version='0.0.0'):
        """Return configuration (not application!) version."""
        return self.get(self.DEFAULT_SECTION_NAME, 'version', version)

    def set_version(self, version='0.0.0', save=True):
        """Set configuration (not application!) version."""
        self.set(self.DEFAULT_SECTION_NAME, 'version', version, save=save)

    def set_logged_data(self, url=None, brand=None):
        """Set `logged_api_url` and `logged_brand` config options."""
        self.set(self.DEFAULT_SECTION_NAME, 'logged_api_url', url)
        self.set(self.DEFAULT_SECTION_NAME, 'logged_brand', brand)

    def get_logged_data(self):
        """Set `logged_brand` and `logged_api_url` config options."""
        return (
            self.get(self.DEFAULT_SECTION_NAME, 'logged_brand'),
            self.get(self.DEFAULT_SECTION_NAME, 'logged_api_url'),
        )

    def is_logged_brand(self, url=None, brand=None):
        """Check if given `url` or `brand` match current repo."""
        logged_brand, logged_url = self.get_logged_data()

        if brand and logged_brand:
            return brand == logged_brand

        if logged_url and url:
            return url == logged_url

        return False

    def load_from_ini(self):
        """Load config from the associated .ini file."""
        try:
            self.read(self.filename(), encoding='utf-8')
        except configparser.MissingSectionHeaderError:
            print('Warning: File contains no section headers.')

    def __load_old_defaults(self, old_version):
        """Read old defaults."""
        old_defaults = configparser.ConfigParser()
        path = os.path.dirname(self.filename())
        path = os.path.join(path, 'defaults')
        old_defaults.read(os.path.join(path, 'defaults-' + old_version + '.ini'))
        return old_defaults

    @staticmethod
    def __save_new_defaults(defaults, new_version, subfolder):
        """Save new defaults."""
        new_defaults = DefaultsConfig(
            name='defaults-' + new_version,
            subfolder=subfolder,
        )
        if not os.path.isfile(new_defaults.filename()):
            new_defaults.set_defaults(defaults)
            new_defaults._save()  # pylint: disable=protected-access

    def __update_defaults(self, defaults, old_version, verbose=False):
        """Update defaults after a change in version."""
        old_defaults = self.__load_old_defaults(old_version)
        for section, options in defaults:
            for option in options:
                new_value = copy.deepcopy(options[option])
                try:
                    old_value = old_defaults.get(section, option)
                except (configparser.NoSectionError, configparser.NoOptionError):
                    old_value = None
                if (old_value is None) or (str(new_value) != old_value):
                    self._set(section, option, new_value, verbose)

    def __remove_deprecated_options(self, old_version):
        """Remove options present in the .ini file but not in defaults."""
        old_defaults = self.__load_old_defaults(old_version)
        for section in old_defaults.sections():
            for option, _ in old_defaults.items(section, raw=self.raw):
                if self.get_default(section, option) is NoDefault:
                    self.remove_option(section, option)
                    if len(self.items(section, raw=self.raw)) == 0:
                        self.remove_section(section)

    def cleanup(self):
        """Remove .ini file associated to config."""
        os.remove(self.filename())

    def set_as_defaults(self):
        """Set defaults from the current config."""
        self.defaults = []
        for section in self.sections():
            secdict = {}
            for option, value in self.items(section, raw=self.raw):
                secdict[option] = value
            self.defaults.append((section, secdict))

    def get_defaults(self):  # pylint: disable=missing-function-docstring
        self.reset_to_defaults(save=False)

        with io.StringIO() as fd:  # pylint: disable=invalid-name
            self.write(fd)
            fd.seek(0)
            return fd.read()

    def reset_to_defaults(self, save=True, verbose=False, section=None):
        """Reset config to Default values."""
        for sec, options in self.defaults:
            if section is None or section == sec:
                for option in options:
                    value = copy.deepcopy(options[option])
                    self._set(sec, option, value, verbose)
        if save:
            self._save()

    def __check_section_option(self, section, option):
        """Private method to check section and option types."""
        if section is None:
            section = self.DEFAULT_SECTION_NAME
        elif not isinstance(section, str):
            raise RuntimeError("Argument 'section' must be a string")
        if not isinstance(option, str):
            raise RuntimeError("Argument 'option' must be a string")
        return section

    def get_default(self, section, option):
        """
        Get Default value for a given (section, option).

        -> useful for type checking in 'get' method
        """
        section = self.__check_section_option(section, option)
        for sec, options in self.defaults:
            if (sec == section) and (option in options):
                return copy.deepcopy(options[option])
        return NoDefault

    def get(self, section, option, default=NoDefault):  # pylint: disable=arguments-differ
        """
        Get an option.

        section=None: attribute a default section name
        default: default value (if not specified, an exception
        will be raised if option doesn't exist)
        """
        section = self.__check_section_option(section, option)

        if not self.has_section(section):
            if default is NoDefault:
                raise configparser.NoSectionError(section)
            self.add_section(section)

        if not self.has_option(section, option):
            if default is NoDefault:
                raise configparser.NoOptionError(option, section)
            self.set(section, option, default)
            return default

        value = configparser.ConfigParser.get(self, section, option, raw=self.raw)
        default_value = self.get_default(section, option)

        if isinstance(default_value, bool):
            value = ast.literal_eval(value)
        elif isinstance(default_value, float):
            value = float(value)
        elif isinstance(default_value, int):
            value = int(value)
        else:
            with contextlib.suppress(BaseException):
                # lists, tuples, ...
                value = ast.literal_eval(value)
        return value

    def set_default(self, section, option, default_value):
        """
        Set Default value for a given (section, option).

        -> called when a new (section, option) is set and no default exists
        """
        section = self.__check_section_option(section, option)
        for sec, options in self.defaults:
            if sec == section:
                options[option] = default_value

    def set(  # pylint: disable=arguments-differ,too-many-arguments
            self, section, option, value, verbose=False, save=True,
    ):
        """
        Set an option.

        section=None: attribute a default section name
        """
        section = self.__check_section_option(section, option)
        default_value = self.get_default(section, option)
        if default_value is NoDefault:
            default_value = value
            self.set_default(section, option, default_value)

        if isinstance(default_value, bool):
            value = bool(value)
        elif isinstance(default_value, float):
            value = float(value)
        elif isinstance(default_value, int):
            value = int(value)
        elif not isinstance(default_value, str):
            value = repr(value)
        self._set(section, option, value, verbose)
        if save:
            self._save()

    def remove_section(self, section):
        """Remove the specified section from the configuration."""
        configparser.ConfigParser.remove_section(self, section)
        self._save()

    def remove_option(self, section, option):
        """Remove the specified option from the specified section."""
        configparser.ConfigParser.remove_option(self, section, option)
        self._save()



# -*- coding: utf-8 -*-

"""Utilities for common validation tasks."""

from __future__ import annotations

__all__ = [
    # item checkers base
    'Base', 'Checker', 'Converter',

    # composite checkers
    'Composite', 'AllOf', 'AnyOf', 'OneOf',

    # basic checkers
    'IsAny', 'IsEmptyMapping', 'OfChoices', 'OfLength', 'OfType',

    # helper iterators
    'iterable_items', 'mapping_items',

    # checkers for collections
    'CollectionChecker', 'EachPair', 'EachValue', 'FindValue',

    # item operations
    'ItemMixin', 'ItemGetter', 'ItemPopper',

    # helper context managers
    'catch_exception', 'validation_to_warning',

    # shortcuts
    'sequence_of', 'IS_NONE', 'IS_BOOL', 'IS_INT', 'IS_FLOAT', 'IS_STR', 'IS_OPT_BOOL', 'IS_OPT_INT', 'IS_OPT_FLOAT',
    'IS_OPT_STR', 'IS_SEQ', 'IS_MAP',
]

import itertools

import abc
import contextlib
import typing

from anaconda_navigator.utils.logs import logger
from . import exceptions


T = typing.TypeVar('T')
KeyT = typing.TypeVar('KeyT', str, int)

EMPTY: typing.Any = object()


# ╠═══════════════════════════════════════════════════════════════════════════════════════════╡ item checkers base ╞═══╣

class Base(metaclass=abc.ABCMeta):
    """Base for all value checkers."""

    __slots__ = ()

    def convert(self, *args: typing.Callable[[typing.Any], typing.Any]) -> 'Converter':
        """
        Chain a conversion after the primary check passes.

        This might also be used for chaining other checks, especially for checkers which return custom value that should
        be validated further.
        """
        return Converter(self, *args)

    def _field(self) -> exceptions.Field:
        """Field, for which value is returned from current value checker."""
        return ()

    def __and__(self: 'Base', other: 'Base') -> 'AllOf':
        """Require both conditions."""
        return AllOf(self, other)

    @abc.abstractmethod
    def __call__(self, value: typing.Any, *, field: exceptions.Field = ()) -> typing.Any:
        """
        Perform check of the value.

        Returns the validated and (if applicable) converted value.
        """

    def __or__(self, other: Base) -> 'AnyOf':
        """Require at least one of conditions."""
        return AnyOf(self, other)

    def __xor__(self, other: Base) -> 'OneOf':
        """Require exactly one of conditions."""
        return OneOf(self, other)


class Checker(Base, metaclass=abc.ABCMeta):
    """Value checker used for value validation."""

    __slots__ = ()

    @abc.abstractmethod
    def _validate(self, value: typing.Any) -> None:
        """Perform validation of the `value`."""

    def __call__(self, value: typing.Any, *, field: exceptions.Field = ()) -> typing.Any:
        """Perform check of the value."""
        with exceptions.ValidationError.with_field(field):
            try:
                self._validate(value)
            except exceptions.ValidationError:
                raise
            except BaseException:
                raise exceptions.ValidationError('unable to validate value') from None
        return value


class Converter(Base):
    """
    Value checker used for execution of chained conversions.

    :param parent: Parent value checker, which value should be modified.
    :param args: Value conversion functions in order they should be executed.
    """

    __slots__ = ('__args', '__parent')

    def __init__(self, parent: Base, *args: typing.Callable[[typing.Any], typing.Any]) -> None:
        """Initialize new :class:`~Converter` instance."""
        while isinstance(parent, Converter):
            args = parent.args + args
            parent = parent.parent

        self.__args: typing.Final[typing.Tuple[typing.Callable[[typing.Any], typing.Any], ...]] = args
        self.__parent: typing.Final[Base] = parent

    @property
    def args(self) -> typing.Tuple[typing.Callable[[typing.Any], typing.Any], ...]:  # noqa: D401
        """Value conversion functions."""
        return self.__args

    @property
    def parent(self) -> Base:  # noqa: D401
        """Parent checker, which value should be converted."""
        return self.__parent

    def _field(self) -> exceptions.Field:
        """Field, for which value is returned from current value checker."""
        return self.parent._field()  # pylint: disable=protected-access

    def __call__(self, value: typing.Any, *, field: exceptions.Field = ()) -> typing.Any:
        """Perform check of the value."""
        with exceptions.ValidationError.with_field(field):
            result: typing.Any = self.parent(value)
            with exceptions.ValidationError.with_field(self.parent._field()):
                arg: typing.Callable[[typing.Any], typing.Any]
                for arg in self.args:
                    try:
                        result = arg(result)
                    except exceptions.ValidationError:
                        raise
                    except BaseException:
                        raise exceptions.ValidationError('unable to convert value') from None
            return result


# ╠═══════════════════════════════════════════════════════════════════════════════════════════╡ composite checkers ╞═══╣

class Composite(Base, metaclass=abc.ABCMeta):  # pylint: disable=too-few-public-methods
    """Base for composite value checkers."""

    __slots__ = ('__children',)

    def __init__(self, *args: Base) -> None:
        """Initialize new :class:`~Composite` instance."""
        children: typing.List[Base] = []

        arg: Base
        for arg in args:
            if type(arg) is type(self):  # pylint: disable=unidiomatic-typecheck
                children.extend(arg.__children)  # type: ignore  # pylint: disable=protected-access
            else:
                children.append(arg)

        self.__children: typing.Final[typing.Tuple[Base, ...]] = tuple(children)

    @property
    def children(self) -> typing.Tuple[Base, ...]:  # noqa: D401
        """Items to aggregate."""
        return self.__children


class AllOf(Composite):
    """
    Composite checker that requires all conditions.

    Returns the raw value.
    """

    __slots__ = ()

    def __call__(self, value: typing.Any, *, field: exceptions.Field = ()) -> typing.Any:
        """Perform check of the value."""
        child: Base
        for child in self.children:
            child(value, field=field)
        return value


class AnyOf(Composite):
    """
    Composite checker that requires at least one condition.

    Returns output from the first condition that passed the validation.
    """

    __slots__ = ()

    def __call__(self, value: typing.Any, *, field: exceptions.Field = ()) -> typing.Any:
        """Perform check of the value."""
        child: Base
        for child in self.children:
            with contextlib.suppress(exceptions.ValidationError):
                return child(value, field=field)
        raise exceptions.ValidationError('invalid value: none of the conditions are met', field=field)


class OneOf(Composite):
    """
    Composite checker that requires only one condition to pass the check.

    Returns output from the condition that passed.
    """

    __slots__ = ()

    def __call__(self, value: typing.Any, *, field: exceptions.Field = ()) -> typing.Any:
        """Perform check of the value."""
        result: typing.Any
        count: int = 0

        child: Base
        for child in self.children:
            with contextlib.suppress(exceptions.ValidationError):
                result = child(value, field=field)
                count += 1

        if count == 0:
            raise exceptions.ValidationError('none of the conditions are met', field=field)
        if count > 1:
            raise exceptions.ValidationError('multiple conditions are met instead of one', field=field)
        return result


# ╠═══════════════════════════════════════════════════════════════════════════════════════════════╡ basic checkers ╞═══╣

class IsAny(Checker):  # pylint: disable=too-few-public-methods
    """Dumb checker that allows any value."""

    __slots__ = ()

    def _validate(self, value: typing.Any) -> None:
        """Perform validation of the `value`."""


class IsEmptyMapping(Checker):  # pylint: disable=too-few-public-methods
    """Make sure that mapping has no items in it."""

    __slots__ = ()

    def _validate(self, value: typing.Any) -> None:
        """Perform validation of the `value`."""
        if value:
            raise exceptions.ValidationError(f'unexpected items: {", ".join(repr(item) for item in value)}')


class OfChoices(Checker):
    """Make sure that the value if one of the available choices."""

    __slots__ = ('__choices',)

    def __init__(self, *args: typing.Any) -> None:
        """Initialize new :class:`~OfChoices` instance."""
        self.__choices: typing.Final[typing.Tuple[typing.Any, ...]] = args

    @property
    def choices(self) -> typing.Tuple[typing.Any, ...]:  # noqa: D401
        """List of available value choices."""
        return self.__choices

    def _validate(self, value: typing.Any) -> None:
        """Perform validation of the `value`."""
        if value in self.choices:
            return

        message: str = ', '.join(repr(item) for item in self.choices)
        raise exceptions.ValidationError(f'must be one of ({message}), not {value!r}')


class OfLength(Checker):
    """Validate any sizeable structure by its length."""

    __slots__ = ('__at_least', '__at_most')

    def __init__(
            self,
            exact: typing.Optional[int] = None,
            *,
            at_least: typing.Optional[int] = None,
            at_most: typing.Optional[int] = None,
    ) -> None:
        """Initialize new :class:`~OfLength` instance."""
        if at_least is None:
            at_least = exact
        if at_most is None:
            at_most = exact
        self.__at_least: typing.Final[typing.Optional[int]] = at_least
        self.__at_most: typing.Final[typing.Optional[int]] = at_most

    @property
    def at_least(self) -> typing.Optional[int]:  # noqa: D401
        """Minimum allowed number of elements."""
        return self.__at_least

    @property
    def at_most(self) -> typing.Optional[int]:  # noqa: D401
        """Maximum allowed number of elements."""
        return self.__at_most

    def _validate(self, value: typing.Any) -> None:
        """Perform validation of the `value`."""
        if (self.at_least is not None) and (len(value) < self.at_least):
            raise exceptions.ValidationError(f'must have at least {self.at_least} items')
        if (self.at_most is not None) and (len(value) > self.at_most):
            raise exceptions.ValidationError(f'must have at most {self.at_most} items')


class OfType(Checker):
    """
    Validate value by its type.

    :code:`None` might also be used as valid value type.
    """

    __slots__ = ('__allow_none', '__targets')

    def __init__(self, *args: typing.Optional[type]) -> None:
        """Initialize new :class:`~OfType` instance."""
        self.__allow_none: typing.Final[bool] = any(arg is None for arg in args)
        self.__targets: typing.Final[typing.Tuple[type, ...]] = tuple(
            arg
            for arg in args
            if arg is not None
        )

    @property
    def allow_none(self) -> bool:  # noqa: D401
        """Value could be :code:`None`."""
        return self.__allow_none

    @property
    def targets(self) -> typing.Tuple[type, ...]:  # noqa: D401
        """List of valid target types."""
        return self.__targets

    def _validate(self, value: typing.Any) -> None:
        """Perform validation of the `value`."""
        value_type: str
        if value is None:
            if self.allow_none:
                return
            value_type = 'None'
        elif isinstance(value, self.targets):
            return
        else:
            value_type = type(value).__name__

        target_types: str = ' | '.join(itertools.chain(
            (getattr(item, '__name__', str(item)) for item in self.targets),
            (['None'] if self.allow_none else []),
        ))
        raise exceptions.ValidationError(f'must be {target_types}, not {value_type}')


# ╠═════════════════════════════════════════════════════════════════════════════════════════════╡ helper iterators ╞═══╣

def iterable_items(source: typing.Iterable[T]) -> typing.Iterator[T]:
    """
    Iterate through items of any iterable.

    For each item - it's index will be used as the name of a field being validated.

    Example:

    >>> with exceptions.ValidationError.with_field('list'):
    ...     for item in iterable_items([1, 2, 'text', 3]):
    ...         if not isinstance(item, int):
    ...             raise exceptions.ValidationError(message='wrong type', field='value')
    Traceback (most recent call last):
        ...
    anaconda_navigator.api.external_apps.exceptions.ValidationError: list.2.value: wrong type
    """
    index: int
    value: T
    for index, value in enumerate(source):
        with exceptions.ValidationError.with_field(index):
            yield value


def mapping_items(source: typing.Mapping[KeyT, T]) -> typing.Iterator[typing.Tuple[KeyT, T]]:
    """
    Iterate through item tuples of a mapping

    For each such tuple - key of the item will be used as the name of a field being validated.

    Example:

    >>> with exceptions.ValidationError.with_field('dict'):
    ...     for k, v in mapping_items({'field_a': 1, 'field_b': 2, 'field_c': 'text', 'field_d': 3}):
    ...         if not isinstance(v, int):
    ...             raise exceptions.ValidationError(message='wrong type', field='value')
    Traceback (most recent call last):
        ...
    anaconda_navigator.api.external_apps.exceptions.ValidationError: dict.field_c.value: wrong type
    """
    key: KeyT
    value: T
    for key, value in source.items():
        with exceptions.ValidationError.with_field(key):
            yield key, value


# ╠═════════════════════════════════════════════════════════════════════════════════════╡ checkers for collections ╞═══╣

class CollectionChecker(Checker, metaclass=abc.ABCMeta):
    """Common base for checkers of different data collections."""

    __slots__ = ()

    def as_converter(
            self,
            target: typing.Callable[[typing.Iterable[typing.Any]], T],
    ) -> typing.Callable[[typing.Any], T]:
        """Use this checker as a converter to new collection."""
        def result(value: typing.Any) -> T:
            return target(self._scan(value))

        return result

    @abc.abstractmethod
    def _scan(self, value: typing.Any) -> typing.Iterator[typing.Any]:
        """Validate and update items from original collections."""

    def _validate(self, value: typing.Any) -> None:
        """Perform validation of the `value`."""
        for _ in self._scan(value):
            pass


class EachPair(CollectionChecker):
    """Process mapping by checking keys and values of items."""

    __slots__ = ('__key', '__value')

    def __init__(self, key: Base, value: Base) -> None:
        """Initialize new :class:`~EachPair` instance."""
        self.__key: typing.Final[Base] = key
        self.__value: typing.Final[Base] = value

    @property
    def key(self) -> Base:  # noqa: D401
        """Validator for mapping keys."""
        return self.__key

    @property
    def value(self) -> Base:  # noqa: D401
        """Validator for mapping values."""
        return self.__value

    def _scan(self, value: typing.Any) -> typing.Iterator[typing.Any]:
        """Validate and update items from original collections."""
        v_key: typing.Any
        v_value: typing.Any
        for v_key, v_value in mapping_items(value):
            yield self.__key(v_key), self.__value(v_value)


class EachValue(CollectionChecker):
    """Process values from any iterable."""

    __slots__ = ('__value',)

    def __init__(self, value: Base) -> None:
        """Initialize new :class:`~EachValue` instance."""
        self.__value: typing.Final[Base] = value

    @property
    def value(self) -> Base:  # noqa: D401
        """Validator for iterable values."""
        return self.__value

    def _scan(self, value: typing.Any) -> typing.Iterator[typing.Any]:
        """Validate and update items from original collections."""
        item: typing.Any
        for item in iterable_items(value):
            yield self.value(item)


class FindValue(Base):
    """Look through any iterable and find first valid value."""

    __slots__ = ('__default', '__value')

    def __init__(self, value: Base, *, default: typing.Any = EMPTY) -> None:
        """Initialize new :class:`~FindValue` instance."""
        self.__default: typing.Final[typing.Any] = default
        self.__value: typing.Final[Base] = value

    @property
    def default(self) -> typing.Any:  # noqa: D401
        """Default value to use if iterable has no valid values at all."""
        return self.__default

    @property
    def value(self) -> Base:  # noqa: D401
        """Validator for iterable values."""
        return self.__value

    def __call__(self, value: typing.Any, *, field: exceptions.Field = ()) -> typing.Any:
        """Perform check of the value."""
        with exceptions.ValidationError.with_field(field):
            item: typing.Any
            for item in iterable_items(value):
                with contextlib.suppress(exceptions.ValidationError):
                    return self.value(item)

        if self.default is not EMPTY:
            return self.default

        raise exceptions.ValidationError('no valid value found', field=field)


# ╠══════════════════════════════════════════════════════════════════════════════════════════════╡ item operations ╞═══╣

class ItemMixin:
    """Common part for item accessors."""

    __slots__ = ('__default', '__item')

    def __init__(self, item: exceptions.Name, *, default: typing.Any = EMPTY) -> None:
        """Initialize new :class:`~ItemMixin` instance."""
        self.__default: typing.Final[typing.Any] = default
        self.__item: typing.Final[exceptions.Name] = item

    @property
    def default(self) -> typing.Any:  # noqa: D401
        """Value to use if the expected item is not present in the collection."""
        return self.__default

    @property
    def item(self) -> exceptions.Name:  # noqa: D401
        """Key or index of an item to retrieve from collection."""
        return self.__item

    def _field(self) -> exceptions.Name:
        """Field, for which value is returned from current value checker."""
        return self.__item


class ItemGetter(ItemMixin, Base):
    """Retrieve item from a mapping."""

    __slots__ = ()

    def __call__(self, value: typing.Any, *, field: exceptions.Field = ()) -> typing.Any:
        """Perform check of the value."""
        try:
            return value[self.item]
        except KeyError:
            if self.default is EMPTY:
                raise exceptions.ValidationError(f'{self.item!r} must be provided', field=field) from None
            return self.default


class ItemPopper(ItemMixin, Base):
    """Retrieve item from a mapping and remove it."""

    __slots__ = ()

    def __call__(self, value: typing.Any, *, field: exceptions.Field = ()) -> typing.Any:
        """Perform check of the value."""
        try:
            return value.pop(self.item)
        except KeyError:
            if self.default is EMPTY:
                raise exceptions.ValidationError(f'{self.item!r} must be provided', field=field) from None
            return self.default


# ╠══════════════════════════════════════════════════════════════════════════════════════╡ helper context managers ╞═══╣

@contextlib.contextmanager
def catch_exception(
        exception_type: typing.Union[
            typing.Type[BaseException],
            typing.Tuple[typing.Type[BaseException], ...],
        ] = BaseException,
        *,
        field: typing.Union[str, int, typing.Iterable[typing.Union[str, int]]] = (),
        message: typing.Optional[str] = None,
) -> typing.Iterator[None]:
    """
    Convert any :exc:`~Exception` to :exc:`~anaconda_navigator.api.external_apps.exceptions.ValidationError`.

    :param exception_type: Convert only this type(s) of :exc:`~BaseException` instances.
    :param message: Custom message to append.

                    If not provided - string representation of original :exc:`~BaseException` will be used.
    :param field: Report exception for a particular field.

    Examples:

    >>> with catch_exception():
    ...     _ = 1 / 0
    Traceback (most recent call last):
        ...
    anaconda_navigator.api.external_apps.exceptions.ValidationError: division by zero

    >>> with catch_exception(ValueError):
    ...     _ = 1 / 0
    Traceback (most recent call last):
        ...
    ZeroDivisionError: division by zero
    """
    try:
        yield
    except exception_type as error:
        custom_message: str
        if message is not None:
            custom_message = message
        elif isinstance(error, exceptions.ValidationError):
            custom_message = error.message
        else:
            custom_message = str(error)

        raise exceptions.ValidationError(custom_message, field=field) from error


@contextlib.contextmanager
def validation_to_warning(path: str = '*configuration*', *, silent: bool = True) -> typing.Iterator[None]:
    """Convert Validation errors into warnings."""
    try:
        yield
    except exceptions.ValidationError as exception:
        logger.warning('%s: %s', path, exception.details)
        if not silent:
            raise


# ╠════════════════════════════════════════════════════════════════════════════════════════════════════╡ shortcuts ╞═══╣

def sequence_of(*types: typing.Optional[type]) -> Base:
    """Validate and normalize a value in form of a sequence, which might be provided as a single item from it."""
    return AnyOf(
        OfType(*types).convert(lambda item: (item,)),
        OfType(typing.Sequence).convert(EachValue(OfType(*types)).as_converter(tuple)),
    )


IS_NONE: Base = OfType(None)
IS_BOOL: Base = OfType(bool)
IS_INT: Base = OfType(int)
IS_FLOAT: Base = OfType(float)
IS_STR: Base = OfType(str)

IS_OPT_BOOL: Base = OfType(bool, None)
IS_OPT_INT: Base = OfType(int, None)
IS_OPT_FLOAT: Base = OfType(float, None)
IS_OPT_STR: Base = OfType(str, None)

IS_SEQ: Base = OfType(typing.Sequence)
IS_MAP: Base = OfType(typing.Mapping)



# -*- coding: utf-8 -*-

"""Utilities to fetch application configurations from the web."""

from __future__ import annotations

__all__ = ['ConfigurationDownloader', 'DOWNLOADER']

import functools
import os
import tempfile
import typing
from urllib import parse
import uuid

from qtpy import QtCore
import yaml

from anaconda_navigator.config import preferences, feature_flags
from anaconda_navigator.utils import download_manager as downloader
from anaconda_navigator.utils.logs import logger
from anaconda_navigator.utils import singletons
from anaconda_navigator.utils import url_utils

if typing.TYPE_CHECKING:
    from anaconda_navigator.config import structures
    from . import config_utils


class ConfigurationDownloader(QtCore.QObject):  # pylint: disable=too-few-public-methods
    """Utility that manages configuration files downloads for application tiles."""

    sig_configuration = QtCore.Signal()
    sig_resource = QtCore.Signal(str)
    _sig_fetch_resource = QtCore.Signal(str, str)

    def __init__(self, sources: 'structures.ApplicationSources') -> None:
        """Initialize new :class:`~ConfigurationDownloader` instance."""
        super().__init__()
        self.__manager: typing.Optional[downloader.Manager] = None
        self.__resources: typing.Final[typing.Dict[str, downloader.Download]] = {}
        self.__sources: typing.Final[structures.ApplicationSources] = sources

        self._sig_fetch_resource.connect(self.__fetch_resource)

    def download(self) -> None:
        """Download configuration files for application tiles."""
        if self.__manager is not None:
            return
        self.__manager = downloader.Manager()

        directory: str = tempfile.gettempdir()
        group: downloader.Group = downloader.Group(*(
            downloader.Download(
                source.url,
            ).extra(
                verify=True, headers=feature_flags.prepare_headers()
            ).into(
                os.path.join(directory, uuid.uuid4().hex),
            ).via(
                downloader.Medium.TARGET,
            ).process(
                functools.partial(self.__process_configuration, path=source.path),
            )
            for source in self.__sources.iter()
        )).require(
            downloader.Require.NONE,
        )
        group.sig_done.connect(self.__done)
        self.__manager.execute(group)

    def __process_configuration(self, download: downloader.Download, path: str) -> None:
        """Prepare fetched configuration file."""
        source: str = download.usable_file

        if os.path.getsize(source) > preferences.BIGGEST_APPLICATION_CONFIGURATION:
            logger.debug('%s: file is too large', download._url)  # pylint: disable=protected-access
            return

        stream: typing.TextIO
        with open(source, 'rt', encoding='utf-8') as stream:
            content: 'config_utils.AppConfiguration' = yaml.safe_load(stream) or {}

        record: 'config_utils.AppRecord'
        for record in content.values():
            try:
                # pylint: disable=protected-access
                record['image_path'] = self._request_resource(parse.urljoin(download._url, record['image_path']))
            except KeyError:
                continue

        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'wt', encoding='utf-8') as stream:
            yaml.safe_dump(content, stream)

    def __done(self) -> None:
        """
        Finish fetching configuration files.

        .. info::

            Icons could still be fetched in background.
        """
        self.__manager = None
        self.sig_configuration.emit()

    def _request_resource(self, url: str) -> str:
        """
        Request external resource located at `url`.

        Returns path to corresponding local resource.
        """
        target: str = os.path.join(self.__sources.root, url_utils.safe_file_name(url))
        self._sig_fetch_resource.emit(url, target)
        return target

    def __fetch_resource(self, url: str, target: str) -> None:
        """Perform resource fetching."""
        if url in self.__resources:
            return

        download: downloader.Download = (
            downloader.Download(url).extra(verify=True).into(target).keep(downloader.Keep.FRESH)
        )
        self.__resources[url] = download
        download.sig_done.connect(self.__process_resource)
        download._worker.start()  # pylint: disable=protected-access

    def __process_resource(self, download: downloader.Download) -> None:
        """Process resource after it is fetched."""
        self.__resources.pop(download._url)  # pylint: disable=protected-access
        if download.status == downloader.Status.SUCCEEDED:
            self.sig_resource.emit(download.usable_file)


DOWNLOADER: typing.Final[singletons.Singleton[ConfigurationDownloader]] = singletons.SingleInstanceOf(
    ConfigurationDownloader,
    sources=preferences.REMOTE_APPLICATION_SOURCES,
)



# -*- coding: utf8 -*-

"""Implementation of a "What's new" dialog."""

from __future__ import annotations

__all__ = ['Update', 'WhatsNewDialog']

import collections.abc
import functools
import io
import typing

import attrs

from qtpy import QtCore
from qtpy import QtWebEngineWidgets
from qtpy import QtWidgets

from anaconda_navigator.static import content as static_content
from anaconda_navigator.config import CONF
from anaconda_navigator import widgets
from anaconda_navigator.widgets import dialogs
from . import common

if typing.TYPE_CHECKING:
    import typing_extensions


def mode() -> typing.Literal['light', 'dark']:
    """Detect the current mode."""
    if CONF.get('main', 'dark_mode'):
        return 'dark'
    return 'light'


def prepare_url(url: str) -> QtCore.QUrl:
    """Prepare URL with injected details on light/dark mode."""
    result: QtCore.QUrl = QtCore.QUrl(url)
    query: QtCore.QUrlQuery = QtCore.QUrlQuery(result.query())
    query.addQueryItem('theme', mode())
    result.setQuery(query)
    return result


class IndexHelper:
    """
    Helper class to manage indexes of visible navigation dots.

    This is needed for cases when we can show fewer dots that we actually need to (e.g. we only have space for 7 dots,
    but we have 11 pages). Works for valid cases as well (i.e. we have enough space for all dots)

    It basically replaces all excessive dots in the middle with a visible dot in the center. This dot is selected when
    any page from the middle is active, and activates the first page of the "middle group" when clicked.

    :param total: The total number of dots that ideally should be rendered.
    :param allowed: The maximum number of dots we can actually draw.
    """

    __slots__ = ('delta', 'left', 'right', 'total')

    def __init__(self, total: int, allowed: int) -> None:
        """Initialize new instance of a :class:`~IndexHelper`."""
        allowed = min(allowed, total)

        self.delta: typing.Final[int] = allowed - total
        self.left: typing.Final[int] = allowed - (allowed // 2) - 1
        self.right: typing.Final[int] = total - (allowed // 2)
        self.total: typing.Final[int] = total

    def __call__(self, index: int) -> int:
        """Retrieve index of a visible dot when :code:`index` page is selected."""
        if index <= self.left:
            return index
        if index < self.right:
            return self.left
        return index + self.delta

    def __iter__(self) -> collections.abc.Iterator[int]:
        """Iterate through page indexes each visible dot represents."""
        yield from range(self.left + 1)
        yield from range(self.right, self.total)


class Update(typing.TypedDict):
    """
    Description of a single update to show.

    .. note::

        Should be retrieved from the server.
    """

    id: str
    title: str
    url: str


@attrs.frozen
class UpdatePage:  # pylint: disable=too-few-public-methods
    """Container for all details related to a single page with updates."""

    info: Update
    page: common.ExtendedWebEnginePage = attrs.field(
        factory=functools.partial(common.ExtendedWebEnginePage, origin='whats-new'),
        init=False,
    )

    def load(self) -> None:
        """Request to load a web page."""
        self.page.load(prepare_url(self.info['url']))


@attrs.frozen
class UpdatePageWithContent(UpdatePage):
    """Custom :class:`~UpdatePage` with a predefined web-page content."""

    content: bytes

    @classmethod
    def from_file(
            cls,
            path: str,
            *,
            title: str = 'What\'s new',
            url: str = 'navigator://content',
            identity: str = ':content:',
    ) -> typing_extensions.Self:
        """Load a web-page from a file."""
        stream: io.BufferedReader
        with open(path, 'rb') as stream:
            content: bytes = stream.read()
        return cls(content=content, info={'id': identity, 'title': title, 'url': url})

    def load(self) -> None:
        """Request to load a web page."""
        self.page.setContent(self.content, 'text/html', prepare_url(self.info['url']))


class WhatsNewDialog(dialogs.DialogBase):  # pylint: disable=too-many-instance-attributes
    """
    Dialog for a "what's new" content.

    :param content: Updates to show in the dialog. If none provided - will fall back to the "whats_new_empty" page.
    """

    sig_navigator_link_requested = QtCore.Signal(object)

    def __init__(self, parent: QtWidgets.QWidget | None = None, content: collections.abc.Iterable[Update] = ()) -> None:
        """Initialize new instance of a :class:`~WhatsNewDialog`."""
        super().__init__(parent)

        pages: list[UpdatePage] = list(map(UpdatePage, content))
        if not pages:
            pages.append(UpdatePageWithContent.from_file(static_content.WHATS_NEW_EMPTY_PAGE))

        self.__pages: typing.Final[tuple[UpdatePage, ...]] = tuple(pages)
        self.__page_index: int = -1
        self.__page_index_helper: typing.Final[IndexHelper] = IndexHelper(len(self.__pages), 7)  # max dots

        self.__pending: bool = False
        self.__ready: bool = False

        # interface

        self.__web_view: typing.Final[QtWebEngineWidgets.QWebEngineView] = QtWebEngineWidgets.QWebEngineView()

        self.__previous_button: typing.Final[QtWidgets.QPushButton] = widgets.ButtonNormal()
        self.__previous_button.setCursor(QtCore.Qt.PointingHandCursor)
        self.__previous_button.setText('< prev')
        self.__previous_button.clicked.connect(lambda: self.__set_page(self.page_index - 1))

        self.__navigation_dots: typing.Final[tuple[widgets.DotControlButton, ...]] = tuple(
            map(self.__create_dot, self.__page_index_helper),
        )

        self.__next_button: typing.Final[QtWidgets.QPushButton] = widgets.ButtonNormal()
        self.__next_button.setCursor(QtCore.Qt.PointingHandCursor)
        self.__next_button.setText('next >')
        self.__next_button.clicked.connect(lambda: self.__set_page(self.page_index + 1))

        control_layout: typing.Final[QtWidgets.QHBoxLayout] = QtWidgets.QHBoxLayout()
        control_layout.setContentsMargins(0, 0, 0, 0)
        control_layout.setSpacing(24)
        control_layout.addWidget(self.__previous_button)
        control_layout.addStretch(1)
        for navigation_dot in self.__navigation_dots:
            control_layout.addWidget(navigation_dot)
        control_layout.addStretch(1)
        control_layout.addWidget(self.__next_button)

        main_layout: typing.Final[QtWidgets.QVBoxLayout] = QtWidgets.QVBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)
        main_layout.setSpacing(0)
        main_layout.addWidget(self.__web_view, 1)
        main_layout.addSpacing(16)
        main_layout.addLayout(control_layout)

        self.label_title_bar.setText('What\'s new')
        self.setFixedSize(600, 800)
        self.setLayout(main_layout)

        # initialize content
        pages[0].page.loadFinished.connect(self.__set_ready)

        page: UpdatePage
        for page in self.__pages:
            page.page.sig_navigator_link_requested.connect(self.sig_navigator_link_requested)
            page.load()

        self.page_index = 0

    @property
    def page_index(self) -> int:  # noqa: D401
        """Index of a current page displayed."""
        return self.__page_index

    @page_index.setter
    def page_index(self, value: int) -> None:
        """Switch to a specific update page."""
        value %= len(self.__pages)

        # update dots
        self.__navigation_dots[self.__page_index_helper(self.__page_index)].setChecked(False)
        self.__navigation_dots[self.__page_index_helper(value)].setChecked(True)

        # skip further processing if page did not change
        if value == self.__page_index:
            return

        # show selected page
        page: UpdatePage = self.__pages[value]
        self.label_title_bar.setText(page.info['title'])
        self.__web_view.setPage(page.page)

        # update buttons
        self.__previous_button.setEnabled(value > 0)
        self.__next_button.setEnabled(value + 1 < len(self.__pages))

        # mark current page as seen
        seen: set[str] = set(CONF.get('internal', 'whats_new_seen'))
        seen.add(page.info['id'])
        CONF.set('internal', 'whats_new_seen', sorted(seen))

        # remember current value
        self.__page_index = value

    def show(self) -> None:
        """Show this dialog to the user."""
        if self.__ready:
            super().show()
        else:
            self.__pending = True

    def __create_dot(self, index: int) -> widgets.DotControlButton:
        """Create a new navigation dot."""
        result: widgets.DotControlButton = widgets.DotControlButton()
        result.clicked.connect(lambda: self.__set_page(index))
        return result

    def __set_page(self, index: int) -> None:
        """
        Switch to an update page.

        .. note::

            Used for lambdas connected to signals.
        """
        self.page_index = index

    def __set_ready(self) -> None:
        """
        Mark a widget as ready to be displayed.

        .. note::

            Should be called after at least a first update page is loaded.
        """
        self.__ready = True
        if self.__pending:
            self.__pending = False
            self.show()



# -*- coding: utf8 -*-

"""Components to manage "what's new" updates."""

from __future__ import annotations

__all__ = ['WhatsNewComponent']

import json
import typing

import attrs.setters
from qtpy import QtCore

from anaconda_navigator import __about__
from anaconda_navigator.api.cloud import login_management
from anaconda_navigator.config import CONF, preferences, feature_flags
from anaconda_navigator.utils import attribution
from anaconda_navigator.utils import download_manager
from anaconda_navigator.widgets.web import whats_new as whats_new_dialogs
from anaconda_navigator.widgets.web import links
from . import common

if typing.TYPE_CHECKING:
    import typing_extensions
    from anaconda_navigator.widgets import main_window


class State(typing.TypedDict, total=False):
    """Local state of what's new updates."""

    cloud_login_popup_state: int
    cloud_login_popup_ts: int


class ClientState(typing.TypedDict):
    """
    State of the current user.

    Used to decide which updates to show.

    .. warning::

        Must not include any PII!
    """

    accounts: list[str]
    navigator_version: str
    state: State


class Selection(typing.TypedDict):
    """
    Selection of updates to show to the user.

    Retrieved from the server.
    """

    state: typing_extensions.NotRequired[State]
    updates: list[whats_new_dialogs.Update]


@attrs.define(on_setattr=attrs.setters.frozen)
class Request:  # pylint: disable=too-few-public-methods
    """
    Details on how to show a "what's new" dialog.

    :param respect_settings: Skip the pop-up if the user selected to "Hide "what's new" dialog on startup".
    :param skip_empty: Skip the pop-up if there are no applicable updates.
    :param updates_only: Skip updates that user already saw.
    """

    respect_settings: bool = attrs.field(default=False, kw_only=True)
    skip_empty: bool = attrs.field(default=False, kw_only=True)
    updates_only: bool = attrs.field(default=False, kw_only=True)

    ready: bool = attrs.field(default=False, init=False, on_setattr=attrs.setters.NO_OP)


class WhatsNewComponent(common.Component):
    """Component to manage what's new content and pop up."""

    __alias__ = 'whats_new'

    def __init__(self, parent: main_window.MainWindow) -> None:
        """Initialize new instance of a :class:`~UpdatesComponent`."""
        super().__init__(parent)

        self.__content: Selection = {'updates': []}
        self.__dialog: whats_new_dialogs.WhatsNewDialog | None = None

        self.__request: Request | None = None
        self.__suspended: bool = True

        self.__timer: QtCore.QTimer = QtCore.QTimer()
        self.__timer.setInterval(3_000)
        self.__timer.setSingleShot(True)
        self.__timer.timeout.connect(self.__push_dialog)

    def fetch(self) -> None:
        """Retrieve latest updates from the server."""
        client_state: ClientState = {
            'accounts': self.main_window.components.accounts.list_accounts(),
            'navigator_version': __about__.__version__,
            'state': {  # type: ignore
                key: CONF.get('internal', key)
                for key in State.__annotations__
            },
        }

        download: download_manager.Download = (
            download_manager.Download(preferences.UPDATES_LINK)
            .attach('POST', json.dumps(client_state))
            .extra(headers={'Content-Type': 'application/json', **feature_flags.prepare_headers()})
            .via(download_manager.Medium.MEMORY)
        )
        download.sig_succeeded.connect(lambda response: self.__process_content(response.content))
        download_manager.MANAGER.instance.execute(download, force=True)

    @typing.overload
    def show(self, request: Request) -> None:
        """Request to show a "what's new" pop up."""

    @typing.overload
    def show(self, *, respect_settings: bool = ..., skip_empty: bool = ..., updates_only: bool = ...) -> None:
        """Request to show a "what's new" pop up."""

    def show(
            self,
            request: Request | None = None,
            *,
            respect_settings: bool = False,
            skip_empty: bool = False,
            updates_only: bool = False,
    ) -> None:
        """
        Request to show a "what's new" pop up.

        Refer to :class:`~Request` for details on arguments.
        """
        if request is None:
            request = Request(respect_settings=respect_settings, skip_empty=skip_empty, updates_only=updates_only)
        if request.respect_settings and CONF.get('main', 'hide_whats_new_dialog'):
            return
        self.__hide_dialog()
        self.__request = request
        self.__timer.start()
        self.fetch()

    def setup(self, worker: typing.Any, output: typing.Any, error: str, initial: bool) -> None:
        """Perform component configuration from `conda_data`."""
        self.__suspended = False
        self.__show_dialog()

    def __process_content(self, content: bytes) -> None:
        """Process response from the server with a new content."""
        try:
            self.__content = json.loads(content)
        except ValueError:
            pass
        self.__push_dialog()

    def __push_dialog(self) -> None:
        """
        Mark requested pop-up as ready to be shown.

        .. note::

            Should be requested after the content is fetched, or after the timeout is reached.
        """
        if self.__request is not None:
            self.__request.ready = True
            self.__show_dialog()

    def __show_dialog(self) -> None:
        """Show requested content in a "what's new" popup."""
        request: Request | None = self.__request
        if self.__suspended or (request is None) or (not request.ready):
            return
        self.__request = None

        # Check if dialog should be skipped due to application settings
        if request.respect_settings and CONF.get('main', 'hide_whats_new_dialog'):
            return

        # Collect updates to include in a dialog
        updates: list[whats_new_dialogs.Update] = self.__content['updates'][:]

        # Skip seen updates if requested
        if request.updates_only:
            seen: set[str] = set(CONF.get('internal', 'whats_new_seen'))

            new_seen: set[str] = set()
            new_updates: list[whats_new_dialogs.Update] = []

            update: whats_new_dialogs.Update
            for update in updates:
                if update['id'] in seen:
                    new_seen.add(update['id'])
                else:
                    new_updates.append(update)

            CONF.set('internal', 'whats_new_seen', sorted(new_seen))
            updates = new_updates

        # Do not show empty dialog if requested
        if request.skip_empty and (not updates):
            return

        self.__dialog = whats_new_dialogs.WhatsNewDialog(content=updates)
        self.__dialog.rejected.connect(self.__hide_dialog)
        self.__dialog.sig_navigator_link_requested.connect(self.__process_navigator_link)
        self.__dialog.show()

        key: str
        state: State = self.__content.pop('state', {})
        for key in State.__annotations__:
            if key in state:
                CONF.set('internal', key, state[key])  # type: ignore

    def __hide_dialog(self) -> None:
        """Hide currently visible "what's new" dialog."""
        if self.__dialog is not None:
            self.__dialog.hide()
            self.__dialog.deleteLater()
            self.__dialog = None

    def __process_navigator_link(self, link: links.NavigatorLink) -> None:
        """
        Process a :code:`navigator://` action triggered from the "what's new" pop-up.

        .. note::

            Implemented on this level to provide access to the main window, other components, and dialog itself (in
            case it should be closed).
        """
        if link.path == ('cloud', 'sign-in'):
            login_management.LOGIN_MANAGER.instance.login(origin='whats-new')
            self.__hide_dialog()
            return

        if link.path == ('cloud', 'sign-up'):
            url: str = attribution.POOL.settings.inject_url_parameters(
                'https://anaconda.cloud/sign-up',
                utm_medium='connect-cloud',
                utm_content='signup',
            )
            links.open_link(url, origin='whats-new')
            return



"""Common helpers for widget signals."""

__all__ = ['DialogOutcome', 'DialogValue']

import enum


class DialogOutcome(enum.Enum):
    """Options for dialog result."""
    REJECT = enum.auto()
    DIALOG_REQUEST = enum.auto()


class DialogValue(enum.Enum):
    """Options for dialog value, that can be selected in the dialog."""

    LOGIN_REMINDER = enum.auto()
    REPO_SELECTOR = enum.auto()



# -*- coding: utf8 -*-

"""Common web components."""

from __future__ import annotations

__all__ = ['ExtendedWebEnginePage']

import typing

from qtpy import QtCore
from qtpy import QtWebEngineWidgets

from . import links


class ExtendedWebEnginePage(QtWebEngineWidgets.QWebEnginePage):
    """Custom :class:`~QWebEnginePage` to open links in default browser."""

    sig_navigator_link_requested = QtCore.Signal(object)

    def __init__(self, *args: typing.Any, origin: str | None = None, **kwargs: typing.Any) -> None:
        """Initialize new instance of :class:`~ExtendedWebEnginePage`."""
        super().__init__(*args, **kwargs)

        self.__origin: typing.Final[str | None] = origin

        self.settings().setAttribute(QtWebEngineWidgets.QWebEngineSettings.ShowScrollBars, False)

    def acceptNavigationRequest(  # pylint: disable=invalid-name
          self,
            url: QtCore.QUrl,
            _type: int,
            isMainFrame: bool,  # pylint: disable=invalid-name
    ) -> bool:
        """Open links in external browser."""
        link: str = url.toString()

        if link.startswith('navigator://'):
            self.sig_navigator_link_requested.emit(links.NavigatorLink.parse(link))
            return False

        if isMainFrame and (_type == QtWebEngineWidgets.QWebEnginePage.NavigationTypeLinkClicked):
            links.open_link(url, origin=self.__origin)
            return False

        return super().acceptNavigationRequest(url, _type, isMainFrame)
